# Copyright 2021 DeepMind Technologies Limited. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""NumPy implementation of the activation transformations used in DKS/TAT."""

import itertools
import os

from absl import logging

from autograd import elementwise_grad as egrad
from autograd import numpy as np
from dks.base.activation_getter import get_activation_function as _get_numpy_activation_function
import scipy.integrate as sp_int
import scipy.optimize as sp_opt
from scipy.special import roots_legendre

# pylint: disable=g-import-not-at-top
# This is a trick to achieve compatibility with multiple versions of SciPy:
try:
  from scipy.integrate._quadrature import _cached_roots_legendre
except ImportError:
  from scipy.integrate.quadrature import _cached_roots_legendre
# pylint: enable=g-import-not-at-top


# Random seed used to initialize activation function parameter searches after
# standard starting points fail (which almost never happens).
_RANDOM_SEED = 123

_CURRENT_DIR = os.path.dirname(os.path.abspath(__file__))

# Feel free to change the path of this file if needed. This file should ideally
# persist between experiments since it can take around 20 minutes to regenerate.
_ROOTS_CACHE_FILE = os.path.join(_CURRENT_DIR, "roots_{}.npy")

_QUADRATURE_ORDER = 100000

# These represent sensible values and ranges to initialize the activation
# function parameter search with. They are sensible because commonly used
# activation functions tend to have reasonable behavior for such values. If
# the searches fail for some activation function they may need to be
# tweaked/expanded.
_ALWAYS_TRY_VALUES = {"input_shift": (0.0, 1.0, -1.0),
                      "input_scale": (1.0, 0.1),
                      "output_shift": (0.0, 1.0, -1.0),
                      "negative_slope": (0.0, 0.5)}

_SAMPLE_RANGE_LOW = {"input_shift": -3.0,
                     "input_scale": 0.0,
                     "output_shift": -3.0,
                     "negative_slope": 0.0}

_SAMPLE_RANGE_HIGH = {"input_shift": 3.0,
                      "input_scale": 2.0,
                      "output_shift": 3.0,
                      "negative_slope": 1.0}


def _precompute_or_load_roots(order):
  """Compute or load the roots used by fixed_quad to save time."""

  if order not in _cached_roots_legendre.cache:

    roots_cache_file = _ROOTS_CACHE_FILE.format(order)

    if os.path.exists(roots_cache_file):

      with open(roots_cache_file, "rb") as fhandle:
        _cached_roots_legendre.cache[order] = np.load(fhandle,
                                                      allow_pickle=False)

    else:
      roots = roots_legendre(order)
      _cached_roots_legendre.cache[order] = roots

      with open(roots_cache_file, "wb") as fhandle:
        np.save(fhandle, roots, allow_pickle=False)


def _estimate_gaussian_mean(fn, order=_QUADRATURE_ORDER):
  """Estimate the mean of a function fn(x) where x ~ N(0,1)."""

  _precompute_or_load_roots(order)

  fn_weighted = lambda x: np.exp(-x**2 / 2) * fn(x)

  integral, _ = sp_int.fixed_quad(fn_weighted, -10., 10., n=order)

  return integral / np.sqrt(2*np.pi)


def _estimate_2d_gaussian_mean(fn):
  """Estimate the mean of a function fn(x, y) where x,y ~ N(0,1)."""

  fn_weighted = lambda x, y: np.exp(-(x**2 + y**2) / 2) * fn(x, y)

  integral, _ = sp_int.dblquad(fn_weighted, -10., 10., -10., 10., epsabs=0.0)

  return integral / (2*np.pi)


def _calc_c_map(activation, derivative_order=0, c=1.0, q_output=None):
  """Evaluate local C map value assuming an input q value of 1.

  Args:
    activation: A callable representing the activation function (applied
      elementwise) from which to define the local C map.
    derivative_order: An integer giving the order of the derivative of the C map
      to take before evaluating it. (Default: 0)
    c: A float giving the input point at which to evaluate the C map. Must
      be between -1.0 and 1.0. (Default: 1.0)
    q_output: Float or None giving the output q value associated with
      ``activation``, if this is known. If None this will be computed from
      scratch. (Default: None)

  Returns:
     A float giving the value of the (derivative of) the local C map for the
     given activation function.
  """

  derivative = activation
  for _ in range(derivative_order):
    derivative = egrad(derivative)

  if c == 0.0:
    integral = _estimate_gaussian_mean(derivative)**2
  elif c == 1.0:
    integral = _estimate_gaussian_mean(lambda x: derivative(x)**2)
  elif c >= -1.0 and c <= 1.0:
    sqrt1mc2 = np.sqrt(1.0 - c**2)
    integral = _estimate_2d_gaussian_mean(
        lambda x, y: derivative(x) * derivative(c*x + sqrt1mc2*y))
  else:
    raise NotImplementedError("Input c value must be between -1.0 and 1.0.")

  if q_output is None:
    q_output = _estimate_gaussian_mean(lambda x: activation(x)**2)

  return integral / q_output


def _calc_c_slope(activation, c=1.0, q_output=None):
  """Evaluate local C map derivative assuming an input q value of 1."""
  return _calc_c_map(activation, derivative_order=1, c=c, q_output=q_output)


def _calc_c_curv(activation, c=1.0, q_output=None):
  """Evaluate local C map second derivative assuming an input q value of 1."""
  return _calc_c_map(activation, derivative_order=2, c=c, q_output=q_output)


def _calc_q_slope_1(activation):
  """Computes the derivative of a local Q map at q=1."""
  derivative = egrad(activation)
  return _estimate_gaussian_mean(lambda x: activation(x)*derivative(x)*x)


def _leaky_relu_cmap(c, negative_slope):
  """Evaluates the local C map for Leaky ReLU with the given negative slope."""
  return ((1 - negative_slope)**2 * (np.sqrt(1 - c**2)
                                     + (np.pi - np.arccos(c)) * c) / np.pi
          + 2 * negative_slope * c) / (1 + negative_slope**2)


def _compute_output_params(act, local_c_val_0_target):
  """Compute output params to achieve Q(1)=1 and C(0)=local_c_val_0_target."""

  if local_c_val_0_target is not None:
    output_shift = np.sqrt(local_c_val_0_target) - _estimate_gaussian_mean(act)
    act_shifted = lambda x: act(x) + output_shift
  else:
    output_shift = None
    act_shifted = act

  output_scale = 1. / np.sqrt(
      _estimate_gaussian_mean(lambda x: act_shifted(x)**2))

  return output_shift, output_scale


def _transform_activation(phi, params):
  """Transform an activation function phi using the given parameters."""

  params = params.copy()

  input_scale = params.pop("input_scale", None)
  input_shift = params.pop("input_shift", None)
  output_shift = params.pop("output_shift", None)
  output_scale = params.pop("output_scale", None)

  def activation(x):

    # Note: DO NOT use += and *= below! Bad things will happen.
    if input_scale is not None:
      x = x * float(input_scale)
    if input_shift is not None:
      x = x + float(input_shift)

    x = phi(x, **params)

    if output_shift is not None:
      x = x + float(output_shift)
    if output_scale is not None:
      x = x * float(output_scale)

    return x

  return activation


def _solve_for_activation_params(
    name, local_q_slope_target, local_c_val_0_target, local_c_slope_1_target,
    local_c_slope_0_target, local_c_curv_target, reject_condition=None,
    num_tries=50):
  """Computes activation function parameters to achieve the given targets."""

  # Making sure random starting points used in solvers will be the same for
  # each run, so that they find the same exact solutions (which is important
  # when using JAX with multiple processes).
  np.random.seed(_RANDOM_SEED)

  # RELU has 1 less degree of freedom so we use this special-case logic:
  if name == "relu":
    # The constant below is the maximum value of local_c_slope_1 that can be
    # achieved in ReLUs given a positive input shift value.
    if local_c_slope_1_target < 1.466942206924260361:
      input_shift = 1.0
    else:
      input_shift = -1.0

    constant_params = {"input_shift": input_shift}
    opt_params = ("input_scale",)

    local_q_slope_target = None  # we turn this off for ReLUs
    assert local_c_slope_0_target is None
    assert local_c_curv_target is None

  elif name == "leaky_relu":
    constant_params = {}
    opt_params = ("negative_slope", "input_shift")

  elif local_c_val_0_target is not None:
    constant_params = {}
    opt_params = ("input_scale", "input_shift")

  else:
    constant_params = {}
    opt_params = ("input_scale", "input_shift", "output_shift")

  def make_params_from_pvector(p):
    """Make a dictionary of activation parameters from a given vector."""

    params = constant_params.copy()

    for i, pname in enumerate(opt_params):
      assert pname not in params
      params[pname] = p[i]

    # We directly compute output_scale, and sometimes also output_shift, instead
    # of taking them from the parameter vector.
    output_shift, output_scale = _compute_output_params(
        _transform_activation(
            _get_numpy_activation_function(name), params),
        local_c_val_0_target)

    if output_shift is not None:
      assert "output_shift" not in params
      params["output_shift"] = output_shift

    if output_scale is not None:
      assert "output_scale" not in params
      params["output_scale"] = output_scale

    return params

  def compute_error_vector(p):
    """Computes vector of errors (value - target) over relevant quantities."""

    params = make_params_from_pvector(p)
    phi_hat = _transform_activation(_get_numpy_activation_function(name),
                                    params)

    residual_array = []

    if local_q_slope_target is not None:
      local_q_slope = _calc_q_slope_1(phi_hat)
      residual_array += [local_q_slope - local_q_slope_target]

    if local_c_slope_1_target is not None:
      local_c_1_slope = _calc_c_slope(phi_hat, c=1.0, q_output=1.0)
      residual_array += [local_c_1_slope - local_c_slope_1_target]

    if local_c_slope_0_target is not None:
      local_c_0_slope = _calc_c_slope(phi_hat, c=0.0, q_output=1.0)
      residual_array += [local_c_0_slope - local_c_slope_0_target]

    if local_c_curv_target is not None:
      local_c_curv = _calc_c_curv(phi_hat, c=1.0, q_output=1.0)
      residual_array += [local_c_curv - local_c_curv_target]

    return np.asarray(residual_array)

  # Make the starting points for the search:
  always_try = tuple(_ALWAYS_TRY_VALUES[pname] for pname in opt_params)
  starting_points = list(itertools.product(*always_try))
  for _ in range(num_tries - len(starting_points)):
    starting_points += [
        tuple(np.random.uniform(low=_SAMPLE_RANGE_LOW[pname],
                                high=_SAMPLE_RANGE_HIGH[pname])
              for pname in opt_params)]

  # For each starting point we run sp_opt.root to try and find a solution:
  for starting_point in starting_points:

    sol = sp_opt.root(compute_error_vector, np.asarray(starting_point),
                      method="hybr", jac=False, options=None)
    if sol.success:
      params = make_params_from_pvector(sol.x)
      if reject_condition is None or not reject_condition(name, params):
        break

    logging.debug("Failed to find parameters from starting point %s.",
                  starting_point)

  if not sol.success:
    raise ValueError(f"Failed to find parameters for '{name}'!")

  logging.info("Found parameters for '%s': %s", name, params)

  return params


def _solve_increasing(fn, target, input_, min_, max_, tol=1e-8, max_eval=100):
  """Solves for x in fn(x)=target, where fn is an increasing function.

  Args:
    fn: A callable which takes a scalar input x and produces a scalar output.
      Must compute an increasing function.
    target: The target output value of ``fn``.
    input_: The initial guess for x.
    min_: A lower bound on the possible value of x.
    max_: An upper bound on the possible value of x.
    tol: A float which giving the acceptable tolerance for ``|fn(x) - target|``.
      (Default: 1e-8)
    max_eval: An integer giving the maximum number of times to evaluate ``fn``
      before giving up. (Default: 100)

  Returns:
    A float giving a value of x such that ``|fn(x) - target| < tol``.
  """

  # The method used to find the solution is a simple binary search in the
  # inteval [min_, max_], where max_ or min_ will change each iteration. If
  # max_ is infinity we double our current guess instead of averaging it with
  # max_.
  for _ in range(max_eval):

    value = fn(input_)

    logging.debug("binary search vals: min = %f, input = %f, max = %f, "
                  "target = %f, value = %f", min_, input_, max_, target, value)

    if np.abs(value - target) < tol:
      return input_

    if value > target:
      max_ = input_
      input_ = 0.5 * (input_ + min_)

    elif value < target:
      min_ = input_

      if np.isinf(max_):
        input_ = input_ * 2
      else:
        input_ = 0.5 * (input_ + max_)

  raise ValueError(f"Maximum evaluations ({max_eval}) exceeded while searching "
                   "for solution. This is probably due the specified target "
                   "being unachievable for the given architecture. For example,"
                   " a Leaky-ReLU MLP of only a few layers may not be able "
                   "to achieve the default C(0) target of 0.9 under TAT. "
                   "Intuitively, this is because a shallow network cannot be "
                   "made sufficiently nonlinear with such activation functions."
                   " The solution to this would be to either use a smaller "
                   "value for the C(0) target (corresponding to a more linear "
                   "model), or to a use a deeper architecture.")


def _compute_local_c_slope_1_target(max_slope_func, target_value):
  return _solve_increasing(max_slope_func, target_value, 1.1, 1., np.inf)


def _compute_local_c_slope_0_target(max_slope_func, target_value):
  return _solve_increasing(max_slope_func, target_value, 0.99, 0., 1.0)


def _compute_local_c_curv_target(max_curv_func, target_value):
  return _solve_increasing(max_curv_func, target_value, 0.1, 0., np.inf)


def _compute_negative_slope_param(max_lrelu_c0, target_value):
  return _solve_increasing(
      lambda a: 1.0 - max_lrelu_c0(a), 1.0 - target_value, 0.5, 0., 1.0)


def _verify_params_dict_and_set_defaults(params_dict, defaults):
  """Verify keys in parameter dict and set any missing ones to the defaults."""

  bad_keys = set(params_dict.keys()).difference(set(defaults.keys()))

  if bad_keys:
    raise ValueError(
        f"Parameter dictionary had unrecognized keys: '{bad_keys}'")

  for key in defaults.keys():
    if key not in params_dict:
      params_dict[key] = defaults[key]


def _get_activations_params(
    activation_names, method="DKS", dks_params=None, tat_params=None,
    max_slope_func=None, max_curv_func=None, subnet_max_func=None):
  """Get dict of optimized parameters for given named activation functions."""

  if not isinstance(activation_names, (list, tuple)):
    raise ValueError("activation_names argument must be a list or tuple of "
                     "strings.")

  # Note that using dictionaries as defaults in the function def is bad, hence
  # we do this instead:
  if dks_params is None:
    dks_params = {}
  if tat_params is None:
    tat_params = {}

  _verify_params_dict_and_set_defaults(
      dks_params, {"c_slope_1_target": 1.5, "local_q_slope_target": 1.0})
  _verify_params_dict_and_set_defaults(
      tat_params, {"c_val_0_target": 0.9, "c_curve_target": 0.3})

  local_q_slope_target = None
  local_c_val_0_target = None
  local_c_slope_1_target = None
  local_c_slope_0_target = None
  local_c_curv_target = None
  reject_condition = None

  if method == "DKS":

    if "relu" in activation_names:
      logging.warning("The use of ReLUs with DKS is *highly* discouraged. You "
                      "are advised to use Leaky ReLUs instead.")

    c_slope_1_target = dks_params["c_slope_1_target"]

    if c_slope_1_target <= 1.0:
      raise ValueError("Invalid value for DKS 'c_slope_1_target' parameter. "
                       "Must be a float greater than 1.0.")

    if max_slope_func is None:
      if subnet_max_func is None:
        raise ValueError("Must supply 'subnet_max_func' if using DKS and not "
                         "passing in 'max_slope_func'.")
      # We can compute the maximal slope function by replacing composition
      # with multiplication in the maximal c value function.
      max_slope_func = lambda x: subnet_max_func(1.0, lambda y: x * y)

    # Three of the four conditions used by DKS. The remaining condition Q(1) = 1
    # is implied.
    local_q_slope_target = dks_params["local_q_slope_target"]
    local_c_val_0_target = 0.0
    # We set the local slope to achieve C'_f(1) <= target over all subnetworks
    # f:
    local_c_slope_1_target = _compute_local_c_slope_1_target(
        max_slope_func, c_slope_1_target)

    logging.info("Found 'local_c_slope_1_target': %s", local_c_slope_1_target)

  elif method == "TAT" and "leaky_relu" not in activation_names:

    if "relu" in activation_names:
      raise ValueError("Standard ReLU is not supported with TAT. Use leaky "
                       "ReLU instead.")

    c_curve_target = tat_params["c_curve_target"]

    if c_curve_target <= 0.0:
      raise ValueError("Invalid value for TAT 'c_curve_target' parameter. Must "
                       "be a float greater than 0.0.")

    if max_curv_func is None:
      if subnet_max_func is None:
        raise ValueError("Must supply 'subnet_max_func' if using TAT with "
                         "smooth activations and not passing in "
                         "'max_curv_func'.")
      # We can compute the maximal curvature function by replacing composition
      # with addition in the maximal c value function.
      max_curv_func = lambda x: subnet_max_func(0.0, lambda y: x + y)

    # Three of the four conditions used by TAT in the smooth case. The remaining
    # condition Q(1) = 1 is implied.
    local_q_slope_target = 1.0
    local_c_slope_1_target = 1.0
    # We set the local second derivative to achieve C''_f(1) <= target over all
    # subnetworks f:
    local_c_curv_target = _compute_local_c_curv_target(
        max_curv_func, c_curve_target)

    logging.info("Found 'local_c_curv_target': %s", local_c_curv_target)

    # This is a hacky fix used to avoid certain 'bad' solutions we observed that
    # seem to have unstable Q maps and higher kernel approximation errors. It
    # should probably be replaced with something more principled.
    reject_condition = lambda name, params: (  # pylint: disable=g-long-lambda
        params["input_scale"] * params["output_scale"] >= 2.0)

  elif method == "TAT" and "leaky_relu" in activation_names:

    if len(activation_names) > 1:
      raise ValueError("When using Leaky ReLU with TAT it must be the only "
                       "activation function.")

    c_val_0_target = tat_params["c_val_0_target"]
    if c_val_0_target > 1.0 or c_val_0_target < 0.0:
      raise ValueError("Invalid value for TAT 'c_val_0_target' parameter. Must "
                       "be a float between 0.0 and 1.0.")

    if subnet_max_func is None:
      raise ValueError("Must supply 'subnet_max_func' if using TAT with Leaky "
                       "ReLU activation functions.")

    max_lrelu_c0 = lambda neg_slope: subnet_max_func(  # pylint: disable=g-long-lambda
        0.0, lambda c: _leaky_relu_cmap(c, neg_slope))

    # We set the negative slope parameter to achieve C_f(0) <= target over all
    # subnetworks f:
    negative_slope = _compute_negative_slope_param(
        max_lrelu_c0, c_val_0_target)

    # This is the value required to achieve Q(1) = 1 for Leaky ReLUs. See the
    # TAT paper for details.
    output_scale = np.sqrt(2.0 / (1.0 + negative_slope**2))

    logging.info("Found parameters for 'leaky_relu': negative_slope = %s, "
                 "output_scale = %s.", negative_slope, output_scale)

    return {"leaky_relu": {"output_scale": output_scale,
                           "negative_slope": negative_slope}}

  else:
    raise ValueError(f"Unrecognized value for argument 'method': {method}")

  params = {}
  for name in activation_names:

    params[name] = _solve_for_activation_params(
        name,
        local_q_slope_target=local_q_slope_target,
        local_c_val_0_target=local_c_val_0_target,
        local_c_slope_1_target=local_c_slope_1_target,
        local_c_slope_0_target=local_c_slope_0_target,
        local_c_curv_target=local_c_curv_target,
        reject_condition=reject_condition)

  return params


def get_transformed_activations(
    activation_names, method="DKS", dks_params=None, tat_params=None,
    max_slope_func=None, max_curv_func=None, subnet_max_func=None,
    activation_getter=_get_numpy_activation_function,
):
  """Gets transformed activation functions using the DKS or TAT method.

  See the DKS paper (https://arxiv.org/abs/2110.01765) and the TAT paper
  (https://openreview.net/forum?id=U0k7XNTiFEq) for details about what these
  are, how they are computed, and what their parameters mean. A procedure to
  compute the "maximal slope function" is given in the section titled "Summary
  of our method" of the DKS paper. Procedures to compute the "maximal curvature
  function", and the "subnet maximizing function", are given in the appendix
  section titled "Additional details and pseudocode for activation function
  transformations" of the TAT paper.

  Note that if you are using the JAX, PyTorch, or TensorFlow frameworks, you
  probably want to be using the version of get_transformed_activations() in the
  corresponding subpackage. (These are basically thin wrappers around this
  function that pass a framework-specific value to the ``activation_getter``
  argument.)

  Args:
    activation_names: An iterable of string names for the activation functions.
      Supported names are the intersection of those supported by
      dks.base.activation_getter.get_activation_function, and those supported
      by the getter passed to the ``activation_getter`` argument (which defaults
      to dks.base.activation_getter.get_activation_function). The built-in
      getters in this package (for each framework) currently support the
      following names: "tanh", "softplus", "leaky_relu", "relu" (not
      recommended; use "leaky_relu" instead), "selu", "elu", "swish", "sigmoid",
      "erf", "bentid", "atan", "asinh", "square", "softsign", "gelu", and
      "gelu_exact".
    method: A string representing the method used to transform the activation
      functions. Can be "DKS", "TAT", or "untransformed". The latter choice
      will return activation functions without any transformations.
      (Default: "DKS")
    dks_params: A dictionary containing the parameters to use for DKS. Keys
      should be a subset of {"c_slope_1_target", "local_q_slope_target"}.
      "c_slope_1_target" gives the target maximal slope value for the network
      (corresponding to "zeta" from the paper), and defaults to 1.5.
      "local_q_slope_target" gives the target value for the local Q map slope
      of each nonlinear layer (which is kept at 1.0 in the paper -- except in
      ablation tests), and defaults to 1.0. If ``dks_params`` is passed as None,
      it defaults to the empty dictionary (so that the parameters will use their
      default values). (Default: None)
    tat_params: A dictionary containing the parameters to use for TAT. Keys
      should be a subset of {"c_val_0_target", "c_curve_target"}.
      "c_val_0_target" gives the maximum value of ``C_f(0)`` over subnetworks f,
      which is used when transforming Leaky ReLUs (and corresponds to "eta" from
      the paper), and defaults to 0.9. "c_curve_target" gives the maximum value
      of ``C''_f(1)``, which is used for all other activation functions (and
      corresponds to "tau" from the paper), and defaults to 0.3. If
      ``tat_params`` is passed as None, it defaults to the empty dictionary (so
      that the parameters will use their default values). (Default: None)
    max_slope_func: A callable which computes the "maximal slope function" of
      the network, as defined in the DKS paper. It should take a single argument
      representing the slope of each local C map at ``c=1``. If this is required
      (i.e. when using DKS) but passed as None, it will be generated using
      ``subnet_max_func`` when possible. (Default: None)
    max_curv_func: A callable which computes the "maximal curvature function" of
      the network, as defined in the TAT paper. It should take a single
      parameter representing the second derivative of each local C map at c=1.
      If this is required (i.e. when using TAT with smooth activation functions)
      but is passed as None, it will be generated using ``subnet_max_func`` when
      possible. (Default: None)
    subnet_max_func: A callable which computes the "subnetwork maximizing
      function" of the network, as defined in the TAT paper (and denoted
      ``M_{f,r}(x)``). It should take two arguments: the input value ``x``, and
      a callable ``r_fn`` which maps a float to a float. This is required when
      using TAT with Leaky ReLUs. (Default: None)
    activation_getter: A callable which takes a string name for an activation
      function and returns the (untransformed) activation function corresponding
      to this name. Defaults to one returning activation functions in NumPy
      (with autograd). Returned transformed activation functions will be based
      on the output of this callable. Other tensor frameworks can be supported
      by changing this argument. See the versions of
      get_transformed_activations() in the ``dks.jax``, ``dks.pytorch``, and
      ``dks.tensorflow`` subpackages.

  Returns:
    A dictionary mapping the activation function names to their corresponding
    transformed activation functions.
  """

  if method == "untransformed":
    return {name: activation_getter(name) for name in activation_names}

  params = _get_activations_params(
      activation_names, method=method, dks_params=dks_params,
      tat_params=tat_params, max_slope_func=max_slope_func,
      max_curv_func=max_curv_func, subnet_max_func=subnet_max_func)

  transformed_acts = {}

  for name in activation_names:
    transformed_acts[name] = _transform_activation(activation_getter(name),
                                                   params[name])
  return transformed_acts



# Copyright 2021 DeepMind Technologies Limited. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Data pre-processing functions for use with DKS/TAT in PyTorch."""

import functools
import operator

from absl import logging
import torch


_prod = lambda x: functools.reduce(operator.mul, x, 1)


def per_location_normalization(x, homog_mode="one", homog_scale=1.0,
                               has_batch_dim=True):
  """Applies Per-Location Normalization (PLN) to a given tensor.

  This function generalizes the idea of PLN from the DKS paper to tensors of
  arbitrary shape. Normalization is done over the last dimension (and only the
  last dimension), so that ``torch.mean(PLN(x)**2, dim=-1, keepdim=True)`` is a
  tensor of ones. Note here that "normalization" does not correspond to making
  the vectors at each location have norm 1. Rather, they will have a squared
  norm given by ``x.size()[-1]``.

  All dimensions, except for the last, and possibly the first (which may be the
  batch dimension), are treated as indexing different "locations", analogous to
  how locations are indexed by the height and width dimensions in convolutional
  layers. The last dimension is always considered the "data" or "feature"
  dimension, analogous to the channels dimension in convolutional layers. For
  models where the dimensions don't have this interpretation, this type of
  preprocessing may not be suitable. (And it's likely that the rest of the
  ``dks`` package, and perhaps even the DKS/TAT method iself, won't be
  applicable either.)

  Before normalization occurs, a homogeneous coordinate may be appended to the
  last dimension of the tensor. If and how this depends on the value
  of the arguements ``homog_mode`` and ``homog_scale``, as described in the
  arguments section. This step is designed to preserve the information that
  would otherwise be lost due to normalization.

  The motivation for PLN is to ensure that the input "q values" to a network are
  always 1, which is a technical requirement of DKS/TAT. While DKS/TAT can often
  work well in practice without PLN, there are situations where using PLN will
  be crucial. In particular, if the input data, or particular samples from it,
  have an extreme scale that deviates from the typical ones seen in CIFAR and
  ImageNet (with the standard preprocessing applied). With CIFAR in particular
  we have observed that some pixels in some images have feature vectors that are
  exactly zero, which can lead to problems when using TAT with leaky ReLUs.

  See the section titled "Uniform q values via Per-Location Normalization" in
  the DKS paper (https://arxiv.org/abs/2110.01765) for a discussion of PLN.

  Args:
    x: A PyTorch tensor representing the input to a network to be normalized. If
      ``x`` has a batch dimension it must be the first one.
    homog_mode: A string indicating whether to append a homogeneous coordinate,
      and how to compute it. Can be ``one``, ``avg_q``, or ``off``. If
      ``one``, the coordinate will have the value 1. If ``avg_q``, it will be
      given by taking the mean squared value of ``x`` across the non-batch axes.
      If ``off`, no homogeneous coordinate will be added. (Default: "one")
    homog_scale: A float used to rescale homogenous coordinate (if used).
      (Default: 1.0)
    has_batch_dim: A boolean specifying whether ``x`` has a batch dimension
      (which will always be its first dimension). Note that many data processing
      pipelines will process training cases one at a time. Unless this is done
      with a singleton leading "dummy" batch dimension (which isn't typical)
      this argument should be set to False. (Default: True)

  Returns:
    A PyTorch tensor which is the result of applying PLN to ``x``, as described
    above.
  """

  def q_val(z, dim):
    return torch.mean(torch.square(z), dim=dim, keepdim=True)

  x_shape = list(x.size())

  if len(x_shape) == 0 and has_batch_dim:  # pylint: disable=g-explicit-length-test
    raise ValueError("x doesn't appear to have a batch dimension.")

  if homog_mode == "avg_q" and ((len(x_shape) <= 2 and has_batch_dim)
                                or (len(x_shape) <= 1 and not has_batch_dim)):  # pylint: disable=g-explicit-length-test

    raise ValueError("homog_mode='avg_q' should not be used for datasets with "
                     "no time/location dimension, as it doesn't offer anything "
                     "beyond what homog_mode='off' would in such cases.")

  if ((len(x_shape) == 1 and has_batch_dim)
      or (len(x_shape) == 0 and not has_batch_dim)):  # pylint: disable=g-explicit-length-test

    x = torch.unsqueeze(x, dim=-1)
    x_shape += [1]

  # the threshold 20 is arbitrary
  if _prod(x_shape[1 if has_batch_dim else 0:]) < 20 and homog_mode == "avg_q":

    logging.warning("Using homog_mode='avg_q' for datasets with few total "
                    "degrees of freedom per batch element (taken over "
                    "time/location dimensions and the data dimension) is "
                    "dangerous. This is because it will remove one degree of "
                    "freedom, and possibly destroy important information. See "
                    "the discussion in the subsection of the DKS paper titled "
                    "'Uniform q values via Per-Location Normalization'.")

  if x_shape[-1] < 20 and homog_mode == "off":

    logging.warning("Using homog_mode='off' for datasets with a small data "
                    "dimension is dangerous. This is because it will remove "
                    "one degree of freedom in this dimension, and possibly "
                    "destroy important information. See the discussion in the "
                    "subsection of the DKS paper titled 'Uniform q values via "
                    "Per-Location Normalization'.")

  if homog_mode == "avg_q":

    homog = torch.sqrt(q_val(x, dim=list(range(1 if has_batch_dim else 0,
                                               len(x_shape)))))

    if has_batch_dim:
      homog = torch.tile(homog, [1] + x_shape[1:-1] + [1])

    else:
      homog = torch.tile(homog, x_shape[:-1] + [1])

  elif homog_mode == "one":
    homog = torch.ones(x_shape[:-1] + [1])

  elif homog_mode == "off":
    homog = None

  else:
    raise ValueError(f"Unrecognized value for homog_mode: {homog_mode}.")

  if homog_scale != 1.0 and homog is not None:
    homog = homog_scale * homog

  if homog is not None:
    x = torch.cat([x, homog], dim=-1)

  x = x / torch.sqrt(q_val(x, dim=-1))

  return x



# Copyright 2021 DeepMind Technologies Limited. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""TF implementation of the activation transformations used in DKS/TAT."""

from dks.base import activation_transform
import tensorflow as tf


def _get_tf_activation_function(name):
  """Get activation function by name in TensorFlow."""

  if name == "bentid":
    return lambda x: (tf.sqrt(tf.square(x) + 1.) - 1.) / 2. + x
  elif name == "erf":
    return tf.math.erf
  elif name == "atan":
    return tf.math.atan
  elif name == "asinh":
    return tf.math.asinh
  elif name == "leaky_relu":
    return lambda x, negative_slope=0.01: tf.nn.leaky_relu(  # pylint: disable=g-long-lambda
        x, alpha=negative_slope)
  elif name == "gelu":
    return lambda x: tf.nn.gelu(x, approximate=True)
  elif name == "gelu_exact":
    return lambda x: tf.nn.gelu(x, approximate=False)
  elif hasattr(tf.nn, name):
    return getattr(tf.nn, name)
  else:
    raise ValueError(f"Unrecognized activation function name '{name}'.")


def get_transformed_activations(*args, **kwargs):
  """See ``dks.base.activation_transform.get_transformed_activations()``."""

  return activation_transform.get_transformed_activations(
      *args, **kwargs, activation_getter=_get_tf_activation_function)



# Copyright 2021 DeepMind Technologies Limited. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""PyTorch implementation of the activation transformations used in DKS/TAT."""

import math

from dks.base import activation_transform
import torch
import torch.nn.functional as tfunc


# PyTorch doesn't seem to currently support the commonly used GELU approximation
# in its public-facing API (as of June 2022) so we implement it here instead.
def _gelu_approx(x):
  return (0.5 * x * (1.0 + torch.tanh(math.sqrt(2.0 / math.pi) *
                                      (x + 0.044715 * torch.pow(x, 3)))))


def _get_pytorch_activation_function(name):
  """Get activation function by name in PyTorch."""

  if name == "bentid":
    return lambda x: (torch.sqrt(torch.square(x) + 1.) - 1.) / 2. + x
  elif name == "gelu":
    return _gelu_approx
  elif name == "gelu_exact":
    return tfunc.gelu
  elif hasattr(tfunc, name):
    return getattr(tfunc, name)
  elif hasattr(torch, name):
    return getattr(torch, name)
  else:
    raise ValueError(f"Unrecognized activation function name '{name}'.")


def get_transformed_activations(*args, **kwargs):
  """See ``dks.base.activation_transform.get_transformed_activations()``."""

  return activation_transform.get_transformed_activations(
      *args, **kwargs, activation_getter=_get_pytorch_activation_function)



# Copyright 2021 DeepMind Technologies Limited. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Parameter sampling functions for use with DKS/TAT in JAX."""

import jax
from jax.config import config as jax_config
import jax.numpy as jnp
import numpy as np


def _get_default_float_dtype():
  if jax_config.jax_enable_x64:
    return jnp.float64
  else:
    return jnp.float32


def _uniform_orthogonal(key, shape, scale=1.0, axis=-1, dtype=None):
  """Samples an orthogonal matrix from the uniform/Haar distribution."""

  # The implementation of this function is essentially copied from
  # hk.initializers.Orthogonal.

  if dtype is None:
    dtype = _get_default_float_dtype()

  if len(shape) < 2:
    raise ValueError("Orthogonal initializer requires at least a 2D shape.")

  n_rows = shape[axis]
  n_cols = np.prod(shape) // n_rows
  matrix_shape = (n_rows, n_cols) if n_rows > n_cols else (n_cols, n_rows)

  norm_dst = jax.random.normal(key, matrix_shape, dtype)

  q_mat, r_mat = jnp.linalg.qr(norm_dst)

  q_mat *= jnp.sign(jnp.diag(r_mat))

  if n_rows < n_cols:
    q_mat = q_mat.T

  q_mat = jnp.reshape(q_mat, (n_rows,) + tuple(np.delete(shape, axis)))
  q_mat = jnp.moveaxis(q_mat, 0, axis)

  return jax.lax.convert_element_type(scale, dtype) * q_mat


def scaled_uniform_orthogonal(key, shape, scale=1.0, axis=-1, delta=True,
                              dtype=None):
  """Initializes fully-connected or conv weights using the SUO distribution.

  Output is similar to that of haiku.initializers.Orthogonal, except that it
  supports Delta initializations, and sampled weights are rescaled by
  ``max(sqrt(out_dim / in_dim), 1)``, so that the layer preserves q values at
  initialization-time (assuming initial biases of zero).

  Note that as with all JAX functions, this is pure and totally stateless
  function, and will produce the exact same output given the same arguments.

  Should be used with a zeros initializer for the bias parameters for DKS/TAT.

  See the "Parameter distributions" section of DKS paper
  (https://arxiv.org/abs/2110.01765) for a discussion of the SUO distribution
  and Delta initializations.

  Args:
    key: A PRNG key used as the random key.
    shape: A list of integers giving the shape of the parameter tensor.
    scale: A float giving an additional scale factor applied on top of the
      standard recaling used in the SUO distribution. This should be left
      at its default value when using DKS/TAT. (Default: 1.0)
    axis: An int giving the axis corresponding to the "output dimension" of
      the parameter tensor. (Default: -1)
    delta: A bool determining whether or not to use a Delta initialization
      (which zeros out all weights except those in the central location of
      convolutional filter banks). (Default: True)
    dtype: a float dtype for the return value. (Default: jnp.float64 if
      jax_enable_x64 is true, otherwise jnp.float32).

  Returns:
    The sampled weights as a JAX ndarray.
  """

  if delta and axis != -1:
    raise ValueError("Invalid axis value for Delta initializations. "
                     "Must be -1.")

  if delta and len(shape) != 2:
    # We assume 'weights' is a filter bank when len(shape) != 2

    # In JAX, conv filter banks have the shape
    # [loc_dim_1, loc_dim_2, in_dim, out_dim]
    in_dim = shape[-2]
    out_dim = shape[-1]

    rescale_factor = np.maximum(np.sqrt(out_dim / in_dim), 1.0)

    nonzero_part = _uniform_orthogonal(
        key, shape[-2:], scale=(rescale_factor * scale), axis=-1, dtype=dtype)

    if any(s % 2 != 1 for s in shape[:-2]):
      raise ValueError("All spatial axes must have odd length for Delta "
                       "initializations.")

    midpoints = tuple((s - 1) // 2 for s in shape[:-2])

    return jnp.zeros(shape, dtype).at[midpoints].set(nonzero_part)

  else:

    in_dim = np.prod(np.delete(shape, axis))
    out_dim = shape[axis]

    rescale_factor = np.maximum(np.sqrt(out_dim / in_dim), 1.0)

    return _uniform_orthogonal(
        key, shape, scale=(rescale_factor * scale), axis=axis, dtype=dtype)





# Copyright 2021 DeepMind Technologies Limited. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Haiku initializers for use with DKS/TAT."""

from dks.jax import parameter_sampling_functions
import haiku as hk


class ScaledUniformOrthogonal(hk.initializers.Initializer):
  """SUO (+ Delta) initializer for fully-connected and convolutional layers.

  Similar to hk.initializers.Orthogonal, except that it supports Delta
  initializations, and sampled weights are rescaled by
  ``max(sqrt(out_dim / in_dim), 1)`` so that the layer preserves q values at
  initialization time (assuming initial biases of zero).

  Should be used with a zeros initializer for the bias parameters for DKS/TAT.

  See "Parameter distributions" section of DKS paper
  (https://arxiv.org/abs/2110.01765) for a discussion of the SUO distribution
  and Delta initializations.
  """

  def __init__(self, scale=1.0, axis=-1, delta=True):
    """Construct a Haiku initializer which uses the SUO distribution.

    Args:
      scale: A float giving an additional scale factor applied on top of the
        standard rescaling used in the SUO distribution. This should be left
        at its default value when using DKS/TAT. (Default: 1.0)
      axis: An int giving the axis corresponding to the "output dimension" of
        the parameter tensor. (Default: -1)
      delta: A bool determining whether or not to use a Delta initialization
        (which zeros out all weights except those in the central location of
        convolutional filter banks). (Default: True)
    """

    if delta and axis != -1:
      raise ValueError("Invalid axis value for Delta initializations. "
                       "Must be -1.")
    self.scale = scale
    self.axis = axis
    self.delta = delta

  def __call__(self, shape, dtype):

    return parameter_sampling_functions.scaled_uniform_orthogonal(
        hk.next_rng_key(), shape, scale=self.scale, axis=self.axis,
        delta=self.delta, dtype=dtype)




# Copyright 2021 DeepMind Technologies Limited. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Defines a getter for NumPy activation functions (with autograd support)."""

from autograd import numpy as np
from autograd import scipy as sp


_SELU_LAMBDA = 1.0507009873554804934193349852946
_SELU_ALPHA = 1.6732632423543772848170429916717


_sigmoid = sp.special.expit
_erf = sp.special.erf


def _elu(x, a=1.0, l=1.0):
  is_neg = x < 0
  is_not_neg = np.logical_not(is_neg)
  return l * (is_neg * a * (np.exp(x) - 1) + is_not_neg * x)


def _bentid(x):
  return (np.sqrt(x**2 + 1.) - 1.) / 2. + x


def _softplus(x):
  """Numerically-stable softplus."""
  return np.log(1. + np.exp(-np.abs(x))) + np.maximum(x, 0)


# aka Silu
def _swish(x):
  return x * _sigmoid(x)


def _leaky_relu(x, negative_slope=0.01):
  is_neg = x < 0
  is_not_neg = np.logical_not(is_neg)
  return negative_slope * is_neg * x + is_not_neg * x


# approximate GELU adapted from official JAX implementation:
def _gelu(x):
  cdf = 0.5 * (1.0 + np.tanh(np.sqrt(2.0 / np.pi) * (x + 0.044715 * (x ** 3))))
  return x * cdf


# exact GELU adapted from official JAX implementation:
def _gelu_exact(x):
  return x * (sp.special.erf(x / np.sqrt(2.0)) + 1) / 2


_ACTIVATION_TABLE = {
    "tanh": np.tanh,
    "sigmoid": _sigmoid,
    "erf": _erf,
    "relu": lambda x: np.maximum(0., x),
    "softplus": _softplus,
    "selu": lambda x: _elu(x, _SELU_ALPHA, _SELU_LAMBDA),
    "elu": _elu,
    "swish": _swish,
    "bentid": _bentid,
    "atan": np.arctan,
    "asinh": np.arcsinh,
    "square": lambda x: x**2,
    "softsign": lambda x: x / (1 + np.abs(x)),
    "leaky_relu": _leaky_relu,
    "gelu": _gelu,
    "gelu_exact": _gelu_exact,
}


def get_activation_function(name):
  if name in _ACTIVATION_TABLE:
    return _ACTIVATION_TABLE[name]
  else:
    raise ValueError(f"Unrecognized activation function name '{name}'.")



# Copyright 2021 DeepMind Technologies Limited. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Parameter sampling functions for use with DKS/TAT in TensorFlow."""

import numpy as np
import tensorflow as tf


def _stateless_uniform_orthogonal(shape, seed, gain=1.0,
                                  dtype=tf.dtypes.float32):
  """Samples an orthogonal matrix from the uniform/Haar distribution."""

  # The implementation of this function is essentially copied from
  # tf.initializers.Orthogonal.

  # Check the shape
  if len(shape) < 2:
    raise ValueError("The tensor to initialize must be "
                     "at least two-dimensional. Received: "
                     f"shape={shape} of rank {len(shape)}.")

  # Flatten the input shape with the last dimension remaining
  # its original shape so it works for conv2d
  num_rows = 1
  for dim in shape[:-1]:
    num_rows *= dim

  num_cols = shape[-1]
  flat_shape = (max(num_cols, num_rows), min(num_cols, num_rows))

  a = tf.random.stateless_normal(flat_shape, seed, dtype=dtype)

  # Compute the qr factorization
  q, r = tf.linalg.qr(a, full_matrices=False)

  # Make q uniform
  d = tf.linalg.tensor_diag_part(r)
  q *= tf.sign(d)

  if num_rows < num_cols:
    q = tf.linalg.matrix_transpose(q)

  return gain * tf.reshape(q, shape)


def stateless_scaled_uniform_orthogonal(
    shape, seed, gain=1.0, delta=True, dtype=tf.dtypes.float32):
  """Initializes fully-connected or conv weights using the SUO distribution.

  Similar to a stateess functional version of tf.initializers.Orthogonal, except
  except that it supports Delta initializations, and sampled weights are
  rescaled by ``max(sqrt(out_dim / in_dim), 1)``, so that the layer preserves
  q values at initialization-time (assuming initial biases of zero).

  Note that this is a stateless function, and will produce the exact same output
  given the same arguments. A stateful random op can be created by passing the
  output of some stateful random op as the ``seed`` argument.

  Should be used with a zeros initializer for the bias parameters for DKS/TAT.

  See the "Parameter distributions" section of DKS paper
  (https://arxiv.org/abs/2110.01765) for a discussion of the SUO distribution
  and Delta initializations.

  Args:
    shape: A list of integers giving the shape of the parameter tensor.
    seed: A shape [2] Tensor, the seed to the random number generator. Must have
      dtype tf.int32 or tf.int64. (When using XLA, only tf.int32 is allowed.)
    gain: A float giving an additional scale factor applied on top of the
      standard recaling used in the SUO distribution. This should be left
      at its default value when using DKS/TAT. (Default: 1.0)
    delta: A bool determining whether or not to use a Delta initialization
      (which zeros out all weights except those in the central location of
      convolutional filter banks). (Default: True)
    dtype: a float dtype for the return value. (Default: jnp.float64 if
      jax_enable_x64 is true, otherwise jnp.float32).

  Returns:
    The sampled weights as a TF Tensor.
  """

  if delta and len(shape) != 2:
    # We assume 'weights' is a filter bank when len(shape) != 2

    # In TensorFlow, conv filter banks have the shape
    # [loc_dim_1, loc_dim_2, in_dim, out_dim]
    in_dim = shape[-2]
    out_dim = shape[-1]

    rescale_factor = np.maximum(np.sqrt(out_dim / in_dim), 1.0)

    nonzero_part = _stateless_uniform_orthogonal(
        shape[-2:], seed, gain=(rescale_factor * gain), dtype=dtype)

    if any(s % 2 != 1 for s in shape[:-2]):
      raise ValueError("All spatial axes must have odd length for Delta "
                       "initializations.")

    midpoints = tuple((s - 1) // 2 for s in shape[:-2])

    return tf.scatter_nd((midpoints,), (nonzero_part,), shape)

  else:

    in_dim = np.prod(shape[:-1])
    out_dim = shape[-1]

    rescale_factor = np.maximum(np.sqrt(out_dim / in_dim), 1.0)

    return _stateless_uniform_orthogonal(
        shape, seed, gain=(rescale_factor * gain), dtype=dtype)



# Copyright 2021 DeepMind Technologies Limited. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""A "modified ResNet model" in Haiku with support for both DKS and TAT."""

import math
from typing import Any, Callable, Mapping, Optional, Sequence, Union

from dks.jax import activation_transform
from dks.jax import haiku_initializers
import haiku as hk
import jax.numpy as jnp


FloatStrOrBool = Union[str, float, bool]

BN_CONFIG = {
    "create_offset": True,
    "create_scale": True,
    "decay_rate": 0.999,
}


class BlockV1(hk.Module):
  """ResNet V1 block with optional bottleneck."""

  def __init__(
      self,
      channels: int,
      stride: Union[int, Sequence[int]],
      use_projection: bool,
      bottleneck: bool,
      use_batch_norm: bool,
      activation: Callable[[jnp.ndarray], jnp.ndarray],
      shortcut_weight: Optional[float],
      w_init: Optional[Any],
      name: Optional[str] = None,
  ):

    super().__init__(name=name)

    self.use_projection = use_projection
    self.use_batch_norm = use_batch_norm
    self.shortcut_weight = shortcut_weight

    if self.use_projection and self.shortcut_weight != 0.0:

      self.proj_conv = hk.Conv2D(
          output_channels=channels,
          kernel_shape=1,
          stride=stride,
          w_init=w_init,
          with_bias=not use_batch_norm,
          padding="SAME",
          name="shortcut_conv")

      if use_batch_norm:
        self.proj_batchnorm = hk.BatchNorm(
            name="shortcut_batchnorm", **BN_CONFIG)

    channel_div = 4 if bottleneck else 1

    conv_0 = hk.Conv2D(
        output_channels=channels // channel_div,
        kernel_shape=1 if bottleneck else 3,
        stride=1 if bottleneck else stride,
        w_init=w_init,
        with_bias=not use_batch_norm,
        padding="SAME",
        name="conv_0")

    conv_1 = hk.Conv2D(
        output_channels=channels // channel_div,
        kernel_shape=3,
        stride=stride if bottleneck else 1,
        w_init=w_init,
        with_bias=not use_batch_norm,
        padding="SAME",
        name="conv_1")

    layers = (conv_0, conv_1)

    if use_batch_norm:

      bn_0 = hk.BatchNorm(name="batchnorm_0", **BN_CONFIG)
      bn_1 = hk.BatchNorm(name="batchnorm_1", **BN_CONFIG)

      bn_layers = (bn_0, bn_1)

    if bottleneck:
      conv_2 = hk.Conv2D(
          output_channels=channels,
          kernel_shape=1,
          stride=1,
          w_init=w_init,
          with_bias=not use_batch_norm,
          padding="SAME",
          name="conv_2")

      layers = layers + (conv_2,)

      if use_batch_norm:
        bn_2 = hk.BatchNorm(name="batchnorm_2", **BN_CONFIG)
        bn_layers += (bn_2,)
        self.bn_layers = bn_layers

    self.layers = layers
    self.activation = activation

  def __call__(self, inputs, is_training, test_local_stats):

    out = shortcut = inputs

    if self.use_projection and self.shortcut_weight != 0.0:

      shortcut = self.proj_conv(shortcut)

      if self.use_batch_norm:
        shortcut = self.proj_batchnorm(shortcut, is_training, test_local_stats)

    for i, conv_i in enumerate(self.layers):

      out = conv_i(out)

      if self.use_batch_norm:
        out = self.bn_layers[i](out, is_training, test_local_stats)

      if i < len(self.layers) - 1:  # Don't apply activation on last layer
        out = self.activation(out)

    if self.shortcut_weight is None:
      return self.activation(out + shortcut)

    elif self.shortcut_weight != 0.0:

      return self.activation(
          math.sqrt(1 - self.shortcut_weight**2) * out +
          self.shortcut_weight * shortcut)
    else:
      return out


class BlockV2(hk.Module):
  """ResNet V2 block with optional bottleneck."""

  def __init__(
      self,
      channels: int,
      stride: Union[int, Sequence[int]],
      use_projection: bool,
      bottleneck: bool,
      use_batch_norm: bool,
      activation: Callable[[jnp.ndarray], jnp.ndarray],
      shortcut_weight: Optional[float],
      w_init: Optional[Any],
      name: Optional[str] = None,
  ):

    super().__init__(name=name)

    self.use_projection = use_projection
    self.use_batch_norm = use_batch_norm
    self.shortcut_weight = shortcut_weight

    if self.use_projection and self.shortcut_weight != 0.0:

      self.proj_conv = hk.Conv2D(
          output_channels=channels,
          kernel_shape=1,
          stride=stride,
          w_init=w_init,
          with_bias=not use_batch_norm,
          padding="SAME",
          name="shortcut_conv")

    channel_div = 4 if bottleneck else 1

    conv_0 = hk.Conv2D(
        output_channels=channels // channel_div,
        kernel_shape=1 if bottleneck else 3,
        stride=1 if bottleneck else stride,
        w_init=w_init,
        with_bias=not use_batch_norm,
        padding="SAME",
        name="conv_0")

    conv_1 = hk.Conv2D(
        output_channels=channels // channel_div,
        kernel_shape=3,
        stride=stride if bottleneck else 1,
        w_init=w_init,
        with_bias=not use_batch_norm,
        padding="SAME",
        name="conv_1")

    layers = (conv_0, conv_1)

    if use_batch_norm:

      bn_0 = hk.BatchNorm(name="batchnorm_0", **BN_CONFIG)
      bn_1 = hk.BatchNorm(name="batchnorm_1", **BN_CONFIG)

      bn_layers = (bn_0, bn_1)

    if bottleneck:

      conv_2 = hk.Conv2D(
          output_channels=channels,
          kernel_shape=1,
          stride=1,
          w_init=w_init,
          with_bias=not use_batch_norm,
          padding="SAME",
          name="conv_2")

      layers = layers + (conv_2,)

      if use_batch_norm:

        bn_2 = hk.BatchNorm(name="batchnorm_2", **BN_CONFIG)
        bn_layers += (bn_2,)

        self.bn_layers = bn_layers

    self.layers = layers
    self.activation = activation

  def __call__(self, inputs, is_training, test_local_stats):
    x = shortcut = inputs

    for i, conv_i in enumerate(self.layers):

      if self.use_batch_norm:
        x = self.bn_layers[i](x, is_training, test_local_stats)

      x = self.activation(x)

      if i == 0 and self.use_projection and self.shortcut_weight != 0.0:
        shortcut = self.proj_conv(x)

      x = conv_i(x)

    if self.shortcut_weight is None:
      return x + shortcut

    elif self.shortcut_weight != 0.0:
      return math.sqrt(
          1 - self.shortcut_weight**2) * x + self.shortcut_weight * shortcut

    else:
      return x


class BlockGroup(hk.Module):
  """Higher level block for ResNet implementation."""

  def __init__(
      self,
      channels: int,
      num_blocks: int,
      stride: Union[int, Sequence[int]],
      resnet_v2: bool,
      bottleneck: bool,
      use_projection: bool,
      use_batch_norm: bool,
      activation: Callable[[jnp.ndarray], jnp.ndarray],
      shortcut_weight: Optional[float],
      w_init: Optional[Any],
      name: Optional[str] = None,
  ):
    super().__init__(name=name)

    block_cls = BlockV2 if resnet_v2 else BlockV1

    self.blocks = []
    for i in range(num_blocks):
      self.blocks.append(
          block_cls(
              channels=channels,
              stride=(1 if i else stride),
              use_projection=(i == 0 and use_projection),
              use_batch_norm=use_batch_norm,
              bottleneck=bottleneck,
              shortcut_weight=shortcut_weight,
              activation=activation,
              w_init=w_init,
              name="block_%d" % (i)))

  def __call__(self, inputs, is_training, test_local_stats):

    out = inputs

    for block in self.blocks:
      out = block(out, is_training, test_local_stats)

    return out


def check_length(length, value, name):

  if len(value) != length:
    raise ValueError(f"`{name}` must be of length 4 not {len(value)}")


class ModifiedResNet(hk.Module):
  """Modified version of an Imagenet ResNet model that supports DKS/TAT."""

  CONFIGS = {
      18: {
          "blocks_per_group": (2, 2, 2, 2),
          "bottleneck": False,
          "channels_per_group": (64, 128, 256, 512),
          "use_projection": (False, True, True, True),
      },
      34: {
          "blocks_per_group": (3, 4, 6, 3),
          "bottleneck": False,
          "channels_per_group": (64, 128, 256, 512),
          "use_projection": (False, True, True, True),
      },
      50: {
          "blocks_per_group": (3, 4, 6, 3),
          "bottleneck": True,
          "channels_per_group": (256, 512, 1024, 2048),
          "use_projection": (True, True, True, True),
      },
      101: {
          "blocks_per_group": (3, 4, 23, 3),
          "bottleneck": True,
          "channels_per_group": (256, 512, 1024, 2048),
          "use_projection": (True, True, True, True),
      },
      152: {
          "blocks_per_group": (3, 8, 36, 3),
          "bottleneck": True,
          "channels_per_group": (256, 512, 1024, 2048),
          "use_projection": (True, True, True, True),
      },
      200: {
          "blocks_per_group": (3, 24, 36, 3),
          "bottleneck": True,
          "channels_per_group": (256, 512, 1024, 2048),
          "use_projection": (True, True, True, True),
      },
  }

  def __init__(
      self,
      num_classes: int,
      depth: int,
      resnet_v2: bool = True,
      use_batch_norm: bool = False,
      shortcut_weight: Optional[float] = 0.0,
      activation_name: str = "softplus",
      w_init: Optional[Any] = haiku_initializers.ScaledUniformOrthogonal(
          delta=True),
      logits_config: Optional[Mapping[str, Any]] = None,
      initial_conv_config: Optional[Mapping[str, FloatStrOrBool]] = None,
      dropout_rate: float = 0.0,
      transformation_method: str = "DKS",
      dks_params: Optional[Mapping[str, FloatStrOrBool]] = None,
      tat_params: Optional[Mapping[str, FloatStrOrBool]] = None,
      name: Optional[str] = None,
  ):
    """Constructs a "modified ResNet model" with support for both DKS and TAT.

    By default, we construct the network *without* normalization layers or
    skip connections (making it a "vanilla network"), initialize the weights
    with the SUO distribution, and use DKS to transform the activation functions
    (which are "softplus" by default). These behaviors, and the option to use
    TAT, are controlled via the contructor arguments.

    This file was adapted from the original Haiku ResNet implementation:
    https://github.com/deepmind/dm-haiku/blob/main/haiku/_src/nets/resnet.py
    It is the end result of applying the rules described in the section titled
    "Summary of our method" in the DKS paper (https://arxiv.org/abs/2110.01765)
    to what is essentially a standard ResNet. See the section titled
    "Application to various modified ResNets" in the DKS paper for more details.
    The only departure from this is that we construct the "maximal C map
    function" instead of the "maximal slope function" (which can be computed
    from the former), which enables support for TAT.


    Args:
      num_classes: The number of classes to classify the inputs into.
      depth: The number of layers.
      resnet_v2: Whether to use the v2 ResNet implementation instead of v1.
        Defaults to ``True``.
      use_batch_norm: Whether to use Batch Normalization (BN). Note that DKS/TAT
        are not compatible with the use of BN. Defaults to ``False``.
      shortcut_weight: The weighting factor of shortcut branch, which must be
        a float between 0 and 1, or None. If not None, the shortcut branch is
        multiplied by ``shortcut_weight``, and the residual branch is multiplied
        by ``residual_weight``, where
                ``shortcut_weight**2 + residual_weight**2 == 1.0``.
        If None, no multiplications are performed (which corresponds to a
        standard ResNet), and compatibility with DKS/TAT is lost. Note that
        setting ``shortcut_weight`` to 0.0 effectively removes the skip
        connections from the network. Defaults to ``0.0``.
      activation_name: String name for activation function. To get TReLU from
        the TAT paper one should set this to ``leaky_relu``, and set
        the ``transformation_method`` argument to ``TAT``. Defaults to
        ``softplus``.
      w_init: Haiku initializer used to initialize the weights.
      logits_config: A dictionary of keyword arguments for the logits layer.
      initial_conv_config: Keyword arguments passed to the constructor of the
        initial :class:`~haiku.Conv2D` module.
      dropout_rate: A float giving the dropout rate for penultimate layer of the
        network (i.e. right before the layer which produces the class logits).
        (Default: 0.0)
      transformation_method: A string representing the method used to transform
        the activation function. Can be ``DKS``, ``TAT``, or ``untransformed``.
        Defaults to ``DKS``.
      dks_params: A dictionary containing the parameters to use for DKS. See
        ``dks.base.activation_transform.get_transformed_activations`` for more
        details. Defaults to ``None``.
      tat_params: A dictionary containing the parameters to use for TAT. See
        ``dks.base.activation_transform.get_transformed_activations`` for more
        details. Defaults to ``None``.
      name: Name of the Sonnet module.
    """
    super().__init__(name=name)

    if shortcut_weight is not None and (shortcut_weight > 1.0
                                        or shortcut_weight < 0.0):
      raise ValueError("Unsupported value for shortcut_weight.")

    if (use_batch_norm and
        (transformation_method == "DKS" or transformation_method == "TAT")):
      raise ValueError("DKS and TAT are not compatible with the use of BN "
                       "layers.")

    if (shortcut_weight is None and
        (transformation_method == "DKS" or transformation_method == "TAT")):
      raise ValueError("Must specify a value for shortcut_weight when using "
                       "DKS or TAT.")

    self.depth = depth
    self.resnet_v2 = resnet_v2
    self.use_batch_norm = use_batch_norm
    self.shortcut_weight = shortcut_weight
    self.activation_name = activation_name
    self.dropout_rate = dropout_rate

    blocks_per_group = ModifiedResNet.CONFIGS[depth]["blocks_per_group"]
    channels_per_group = ModifiedResNet.CONFIGS[depth]["channels_per_group"]
    bottleneck = ModifiedResNet.CONFIGS[depth]["bottleneck"]
    use_projection = ModifiedResNet.CONFIGS[depth]["use_projection"]

    logits_config = dict(logits_config or {})
    logits_config.setdefault("w_init", w_init)
    logits_config.setdefault("name", "logits")

    # Number of blocks in each group for ResNet.
    check_length(4, blocks_per_group, "blocks_per_group")
    check_length(4, channels_per_group, "channels_per_group")

    initial_conv_config = dict(initial_conv_config or {})
    initial_conv_config.setdefault("output_channels", 64)
    initial_conv_config.setdefault("kernel_shape", 7)
    initial_conv_config.setdefault("stride", 2)
    initial_conv_config.setdefault("with_bias", not use_batch_norm)
    initial_conv_config.setdefault("padding", "SAME")
    initial_conv_config.setdefault("name", "initial_conv")
    initial_conv_config.setdefault("w_init", w_init)

    act_dict = activation_transform.get_transformed_activations(
        [self.activation_name], method=transformation_method,
        dks_params=dks_params, tat_params=tat_params,
        subnet_max_func=self.subnet_max_func)

    self.activation = act_dict[self.activation_name]

    self.initial_conv = hk.Conv2D(**initial_conv_config)

    if not self.resnet_v2 and use_batch_norm:
      self.initial_batchnorm = hk.BatchNorm(
          name="initial_batchnorm", **BN_CONFIG)

    self.block_groups = []
    strides = (1, 2, 2, 2)
    for i in range(4):
      self.block_groups.append(
          BlockGroup(
              channels=channels_per_group[i],
              num_blocks=blocks_per_group[i],
              stride=strides[i],
              resnet_v2=resnet_v2,
              bottleneck=bottleneck,
              use_batch_norm=use_batch_norm,
              use_projection=use_projection[i],
              shortcut_weight=shortcut_weight,
              activation=self.activation,
              w_init=w_init,
              name="block_group_%d" % (i)))

    if self.resnet_v2 and use_batch_norm:
      self.final_batchnorm = hk.BatchNorm(name="final_batchnorm", **BN_CONFIG)

    self.logits = hk.Linear(num_classes, **logits_config)

  def __call__(self, inputs, is_training, test_local_stats=False):

    out = inputs
    out = self.initial_conv(out)

    if not self.resnet_v2:

      if self.use_batch_norm:
        out = self.initial_batchnorm(out, is_training, test_local_stats)

      out = self.activation(out)

    out = hk.max_pool(
        out, window_shape=(1, 3, 3, 1), strides=(1, 2, 2, 1), padding="SAME")

    for block_group in self.block_groups:
      out = block_group(out, is_training, test_local_stats)

    if self.resnet_v2:
      if self.use_batch_norm:
        out = self.final_batchnorm(out, is_training, test_local_stats)

      out = self.activation(out)

    out = jnp.mean(out, axis=(1, 2))

    if self.dropout_rate > 0.0 and is_training:
      out = hk.dropout(hk.next_rng_key(), self.dropout_rate, out)

    return self.logits(out)

  def subnet_max_func(self, x, r_fn):
    return subnet_max_func(x, r_fn, self.depth, self.shortcut_weight,
                           resnet_v2=self.resnet_v2)


def subnet_max_func(x, r_fn, depth, shortcut_weight, resnet_v2=True):
  """The subnetwork maximizing function of the modified ResNet model."""

  # See Appendix B of the TAT paper for a step-by-step procedure for how
  # to compute this function for different architectures.

  blocks_per_group = ModifiedResNet.CONFIGS[depth]["blocks_per_group"]
  bottleneck = ModifiedResNet.CONFIGS[depth]["bottleneck"]
  use_projection = ModifiedResNet.CONFIGS[depth]["use_projection"]

  if bottleneck and resnet_v2:
    res_fn = lambda z: r_fn(r_fn(r_fn(z)))

  elif (not bottleneck and resnet_v2) or (bottleneck and not resnet_v2):
    res_fn = lambda z: r_fn(r_fn(z))

  else:
    res_fn = r_fn

  res_branch_subnetwork = res_fn(x)

  for i in range(4):
    for j in range(blocks_per_group[i]):

      res_x = res_fn(x)

      if j == 0 and use_projection[i] and resnet_v2:
        shortcut_x = r_fn(x)
      else:
        shortcut_x = x

      x = (shortcut_weight**2 * shortcut_x + (1.0 - shortcut_weight**2) * res_x)

      if not resnet_v2:
        x = r_fn(x)

  x = r_fn(x)

  return max(x, res_branch_subnetwork)



# Copyright 2021 DeepMind Technologies Limited. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Parameter sampling functions for use with DKS/TAT in PyTorch."""

import numpy as np
import torch


def scaled_uniform_orthogonal_(weights, gain=1.0, delta=True):
  """Initializes fully-connected or conv weights using the SUO distribution.

  Similar to torch.nn.init.orthogonal_, except that it supports Delta
  initializations, and sampled weights are rescaled by
  ``max(sqrt(out_dim / in_dim), 1)``, so that the layer preserves q values at
  initialization-time (assuming initial biases of zero).

  Note that as with all PyTorch functions ending with '_', this function
  modifies the value of its tensor argument in-place.

  Should be used with a zeros initializer for the bias parameters for DKS/TAT.

  See the "Parameter distributions" section of DKS paper
  (https://arxiv.org/abs/2110.01765) for a discussion of the SUO distribution
  and Delta initializations.

  Args:
    weights: A PyTorch Tensor corresponding to the weights to be randomly
      initialized.
    gain: A float giving an additional scale factor applied on top of the
      standard recaling used in the SUO distribution. This should be left
      at its default value when using DKS/TAT. (Default: 1.0)
    delta: A bool determining whether or not to use a Delta initialization
      (which zeros out all weights except those in the central location of
      convolutional filter banks). (Default: True)

  Returns:
    The ``weights`` argument (whose value will be initialized).
  """

  shape = list(weights.size())

  if delta and len(shape) != 2:
    # We assume 'weights' is a filter bank when len(shape) != 2

    # In PyTorch, conv filter banks have that shape
    # [in_dim, out_dim, loc_dim_1, loc_dim_2]
    in_dim = shape[0]
    out_dim = shape[1]

    rescale_factor = np.maximum(np.sqrt(out_dim / in_dim), 1.0)

    nonzero_part = torch.nn.init.orthogonal_(weights.new_empty(in_dim, out_dim),
                                             gain=(rescale_factor * gain))

    if any(s % 2 != 1 for s in shape[2:]):
      raise ValueError("All spatial axes must have odd length for Delta "
                       "initializations.")

    midpoints = [(s - 1) // 2 for s in shape[2:]]
    indices = [slice(None), slice(None)] + midpoints

    with torch.no_grad():
      weights.fill_(0.0)
      weights.__setitem__(indices, nonzero_part)

      return weights

  else:

    # torch.nn.orthogonal_ flattens dimensions [1:] instead of [:-1], which is
    # the opposite of what we want here. So we'll first compute the version with
    # the first two dimensions swapped, and then we'll transpose at the end.

    shape = [shape[1], shape[0]] + shape[2:]

    in_dim = np.prod(shape[1:])
    out_dim = shape[0]

    rescale_factor = np.maximum(np.sqrt(out_dim / in_dim), 1.0)

    weights_t = torch.nn.init.orthogonal_(weights.new_empty(shape),
                                          gain=(rescale_factor * gain))
    with torch.no_grad():
      return weights.copy_(weights_t.transpose_(0, 1))
